#----------------------------------------------------------
# SSC_calc.pyx
#
#  Calculate synchrotron radiation and SSC radiation 
#  based on Finke et al. 2008
# Written by R. Itoh
#
# 2012/01/10 v1.0  R.Itoh    New program
# 2016/10/21 v2.0  R.Itoh    Change integral method, class style
#                            Fitting method: pyROOT -> scipy.optimize
# 2018/04/16 v3.0  M.Uemura  Change intgral method from simps to quad
#                            Using cython
#                            Cython usage:
#                            $ cython SSC_calc_cy.pyx
#                            $ python setup.py build_ext --inplace
# 2019/11/23 v4.0  M.Uemura  Delete functions not used in yysscfit
#
# class calc_z(): calculate the distance from redshift based on Wright et al. 2006
# class SSC();    calculate SSC radiation based on Finke et al. 2008 (need calc_z())
#----------------------------------------------------------

import sys,time
import math,numpy
import scipy
import scipy.integrate
import scipy.optimize
cimport numpy
cimport cython

DTYPE = numpy.float32
ctypedef numpy.float32_t DTYPE_t

#######################################################################
#######################################################################
##  CLASS Calc_z  #####################################################
#######################################################################
#######################################################################

class Calc_z():
    
    # Based on Wright (2006, PASP, 118, 1711). 
    # Original Python code at http://www.astro.ucla.edu/~wright/CosmoCalc.html
    
    def __init__(self):
        # initialize constants
        self.verbose  = False
        self.H0       = 75         # Hubble constant
        self.WM       = 0.3        # Omega(matter)
        self.WR       = 0.         # Omega(radiation)
        self.WK       = 0.         # Omega curvaturve = 1-Omega(total)
        self.c        = 299792.458 # velocity of light in km/sec
        self.Tyr      = 977.8      # coefficent for converting 1/H into Gyr
        self.n        = 1000       # number of points in integrals
        
    def calc_distance(self,z):
        # Omega(vacuum) or lambda        
        self.WV       = 1.0 - self.WM - 0.4165/(self.H0*self.H0)
        self.h        = self.H0/100.
        # includes 3 massless neutrino species, T0 = 2.72528
        self.WR       = 4.165E-5/(self.h*self.h)
        self.WK       = 1-self.WM-self.WR-self.WV
        az       = 1.0/(1+1.0*z)
        age      = 0.0
        DTT      = 0.0
        DCMR     = 0.0        
        for i in range(self.n):
            a    = az*(i+0.5)/self.n
            adot = math.sqrt(self.WK+(self.WM/a)+(self.WR/(a*a))+(self.WV*a*a))
            age  = age + 1./adot
        zage     = az*age/self.n
        zage_Gyr = (self.Tyr/self.H0)*zage
        # do integral over a = 1/(1+z) from az to 1 in n steps, midpoint rule
        for i in range(self.n):
            a          = az+(1-az)*(i+0.5)/self.n
            adot       = math.sqrt(self.WK+(self.WM/a)+(self.WR/(a*a))+
                                   (self.WV*a*a))
            DTT        = DTT + 1./adot
            DCMR       = DCMR + 1./(a*adot)
        DTT      = (1.-az)*DTT/self.n
        DCMR     = (1.-az)*DCMR/self.n
        age      = DTT+zage
        age_Gyr  = age*(self.Tyr/self.H0)
        DTT_Gyr  = (self.Tyr/self.H0)*DTT
        DCMR_Gyr = (self.Tyr/self.H0)*DCMR
        DCMR_Mpc = (self.c/self.H0)*DCMR
        # tangential comoving distance
        ratio    = 1.00
        x        = math.sqrt(abs(self.WK))*DCMR
        if x > 0.1:
            if self.WK > 0:
                ratio      = 0.5*(math.exp(x)-math.exp(-x))/x 
            else:
                ratio      = math.sin(x)/x
        else:
            y            = x*x
        if self.WK < 0: y = -y
        ratio          = 1. + y/6. + y*y/120.
        DCMT           = ratio*DCMR
        DA             = az*DCMT
        DA_Mpc         = (self.c/self.H0)*DA
        kpc_DA         = DA_Mpc/206.264806
        DA_Gyr         = (self.Tyr/self.H0)*DA
        DL             = DA/(az*az)
        DL_Mpc         = (self.c/self.H0)*DL
        DL_Gyr         = (self.Tyr/self.H0)*DL

        # comoving volume computation
        ratio = 1.00
        x              = math.sqrt(abs(self.WK))*DCMR
        if x > 0.1:
            if self.WK > 0:
                ratio      = (0.125*(math.exp(2.*x)-math.exp(-2.*x))-x/2.)/\
                             (x*x*x/3.)
            else:
                ratio      = (x/2. - math.sin(2.*x)/4.)/(x*x*x/3.)
        else:
            y            = x*x
        if self.WK < 0: y = -y
        ratio        = 1. + y/5. + (2./105.)*y*y
        VCM          = ratio*DCMR*DCMR*DCMR/3.
        V_Gpc        = 4.*math.pi*((0.001*self.c/self.H0)**3)*VCM

        if self.verbose:
            print 'For H_o = ' + '%1.1f' % self.H0 + ', Omega_M = ' + '%1.2f' % self.WM + ', Omega_vac = ',
            print '%1.2f' % self.WV + ', z = ' + '%1.3f' % z
            print 'It is now ' + '%1.1f' % age_Gyr + ' Gyr since the Big Bang.'
            print 'The age at redshift z was ' + '%1.1f' % zage_Gyr + ' Gyr.'
            print 'The light travel time was ' + '%1.1f' % DTT_Gyr + ' Gyr.'
            print 'The comoving radial distance, which goes into Hubbles law, is',
            print '%1.1f' % DCMR_Mpc + ' Mpc or ' + '%1.1f' % DCMR_Gyr + ' Gly.'
            print 'The comoving volume within redshift z is ' + '%1.1f' % V_Gpc + ' Gpc^3.'
            print 'The angular size distance D_A is ' + '%1.1f' % DA_Mpc + ' Mpc or',
            print '%1.1f' % DA_Gyr + ' Gly.'
            print 'This gives a scale of ' + '%.2f' % kpc_DA + ' kpc/".'
            print 'The luminosity distance D_L is ' + '%1.1f' % DL_Mpc + ' Mpc or ' + '%1.1f' % DL_Gyr + ' Gly.'
            print 'The distance modulus, m-M, is '+'%1.2f' % (5*math.log10(DL_Mpc*1e6)-5)
        self.Ldis = DL_Mpc
        self.sc   = kpc_DA

    def Mpc2cm(self):
        return self.Ldis*1e6*3.08568025e16*1e2


################################################################
################################################################
##  CLASS SSC  #################################################
################################################################
################################################################

cdef class MySED:
    cdef public double v,sed
    def __init__(self, double v, double sed):
        self.v = v
        self.sed = sed

cdef class SSC:
    cpdef public double c,me,h,e,re,sigma,H0,Omega_m,Omega_r,v_step,ysyn_step,vssc_step,yssc_step,z,B,D,T,K,CL,P0,P1,r1,r2,d,cnst_b1,cnst_v2e,cnst_C1,logCL,e0,enessc,eneS,vrange_sy0,vrange_sy1
    cpdef char* tout
    cpdef public object sync
    cpdef public int nsync
    
    # Based on Finke et al. 2008
    # Written by R. Itoh
    
    def __init__(self):
        self.c         = 2.97792458e10  # light speed [cm s^-1]
        self.me        = 9.10938291e-28 # mass of electron [g] 
        self.h         = 6.625e-27      # planc constant [erg s]
        self.e         = 4.80320425e-10 # Fundamental charge in cgs system (statC) 
        self.re        = 2.82e-13       # classical electron radius [cm] 
        self.sigma     = 8.0*math.pi*self.re**2.0/3.0
        self.H0        = 75.0       # Hubble constant
        self.Omega_m   = 0.3        # Omega(matter) 
        self.Omega_r   = 0.0        # Omega(radiation) 
        self.v_step    = 0.2        # step size for frequancy    (log scale)
        self.ysyn_step = 0.1        # step size for SyncEm()     (log scale) 
        self.vssc_step = 0.08      # step size for diff_SSC_v() (log scale)
        self.yssc_step = 0.05       # step size for diff_SSC_r() (log scale)
        self.vrange_sy0 = 1e13
        self.vrange_sy1 = 1e20
        self.tout      = "SSC_output.txt"
        
    cpdef void set_parameters(self, double z0, double B0, double D0, double T0, double K0, double CL0, double P0, double P1, double r1, double r2):
        self.z  = z0  # red shift
        self.B  = B0  # Magnetic field [G]
        self.D  = D0  # Dopplar factor of Bulk emission region
        self.T  = T0  # minimum Time scale of variability [s]
        self.K  = K0  # Normarization of electron distribution
        self.CL = CL0 # Cutoff Lorentz factor
        self.P0 = P0  # Spectral index of electron (Low energy)
        self.P1 = P1  # Spectral index of electron (High energy)
        self.r1 = r1  # Mimimum Lorentz factor
        self.r2 = r2  # Maximum Lorentz factor
        cz      = Calc_z()
        cz.H0   = self.H0
        cz.WM   = self.Omega_m
        cz.calc_distance(self.z)
        self.d  =  cz.Mpc2cm()
        self.cnst_b1 = 4.0*math.pi*self.me**2*self.c**3/(3.0*self.e*self.B*self.h)
        self.cnst_v2e = self.h/(self.me*self.c**2)*(1.0+self.z)/self.D
        self.cnst_C1 = math.sqrt(3.0)*self.D**4*self.e**3*self.B/\
               (4.0*math.pi*self.h*self.d**2)
        self.logCL = numpy.log10(self.CL)
        cdef numpy.ndarray[DTYPE_t,ndim=1] vlst = self.make_log_map(self.vrange_sy0,self.vrange_sy1,self.vssc_step)
        ### Python 2.x
        # self.sync = map(self.SyncEm, vlst)
        ### Python 3
        self.sync = list(map(self.SyncEm, vlst))
        self.nsync = len(self.sync)

    cdef double FC(self,double Q,double TT,double yy):
        cdef double FFCC
        if 1.0/(4.0*yy**2) <= Q and Q <= 1:
            FFCC = (2.0*Q*math.log(Q,math.e) + (1.0+2.0*Q)*(1.0-Q) \
                + 0.5*pow((TT*Q),2.0)/(1.0+TT*Q)*(1.0-Q)) 
        else:
            FFCC = 0
        return FFCC

    cdef double HC(self,double Val):
        cdef double HCC
        if Val < 0.0:HCC = 0.0
        elif Val > 0.0:HCC = 1.0
        else:HCC = 0.5 
        return HCC

    cdef double eleDis(self,double y_ene):
        cdef double ne
        ne = self.K*( pow(y_ene/self.CL,-1.0*self.P0)*self.HC(self.CL - y_ene) + \
                      pow(y_ene/self.CL,-1.0*self.P1)*self.HC( y_ene - self.CL ) )*\
                      self.HC(y_ene - self.r1)*self.HC( self.r2 - y_ene)
        return ne

    cdef double RF(self,double X1):
        cdef double RX=0.0,A0,A1,A2,A3,A4,A5,Y1,lRX
        # R(x) = x/2 \int_0^{pi} d\theta \sin theta \int_{x/\sin\theta}^{\infty} dt K_{5/3}(t) 
        # log(R) = A0 + A1*y + A2*y^2 + A3*y^3 + A4*y^4 + A5*y^5 
        if X1 <= 1e-2:
            RX = 1.80842*pow(X1,1.0/3.0)
        elif 1e-2 < X1 and X1 <= 1.0:
            A0  = -0.35775237
            A1  = -0.83695385
            A2  = -1.1449608
            A3  = -0.68137283
            A4  = -0.22754737
            A5  = -0.031967334
            Y1  = math.log(X1,10.0) 
            lRX = A0 + A1*Y1 + A2*Y1**2 + A3*Y1**3 + A4*Y1**4 + A5*Y1**5 
            RX  = pow(10,lRX)
        elif 1.0 < X1 and X1 <= 10.0:
            A0  = -0.35842494
            A1  = -0.79652041
            A2  = -1.6113032
            A3  =  0.26055213
            A4  = -1.6979017
            A5  =  0.032955035
            Y1  = math.log(X1,10.0) 
            lRX = A0 + A1*Y1 + A2*Y1**2 + A3*Y1**3 + A4*Y1**4 + A5*Y1**5 
            RX  = pow(10,lRX)
        elif 10.0 < X1 :
            RX = math.pi/2.0*math.exp(-1*X1)*(1.0 - 99.0/162.0/X1)
        return RX

    cdef make_log_map(self,double r0,double r1,double rstep):
        cdef double lr0 = numpy.log10(r0)
        cdef double lr1 = numpy.log10(r1)
        cdef numpy.ndarray[DTYPE_t,ndim=1] rlst = numpy.arange(lr0,lr1,rstep,dtype=DTYPE)
        return 10**rlst

    ########################################################
    ##  Synchrotron  #######################################
    ########################################################
    
    cpdef double diff_Sync(self,double logX):
        cdef double X = 10**logX
        cdef double b1 = self.cnst_b1*self.eneS/(X**2)
        cdef double result = self.eleDis(X)*self.RF(b1)*X*numpy.log(10)
        return result

    cpdef double SyncEm(self, double inputv):
        self.eneS = self.cnst_v2e*inputv
        cdef double C1    = self.cnst_C1*self.eneS
        cdef double sync   = C1*(scipy.integrate.quad(self.diff_Sync,0.0,self.logCL,epsabs=0.0,epsrel=1e-4)[0]+scipy.integrate.quad(self.diff_Sync,self.logCL,10,epsabs=0.0,epsrel=1e-4)[0])
        return sync

    ########################################################
    ##  SSC  ###############################################
    ########################################################

    cpdef double diff_SSC_r(self,double logy):
        cdef double y = 10**logy
        cdef double T = 4.0*self.e0*y
        cdef double q = 0.0
        if  1.0-self.enessc/y != 0.0:
            q = (self.enessc/y)/(T*(1.0-(self.enessc/y)))
        else:q = 0.0
        cdef double result = self.eleDis(y)/y**2*self.FC(q,T,y)*y*numpy.log(10)
        return result

    cpdef double diff_SSC_v(self, MySED d):
        cdef double vsync = d.v
        cdef double fsync = d.sed
        self.e0  = self.cnst_v2e*vsync
        cdef double ymin     = 0.5*self.enessc*(1.0+math.sqrt(1+1.0/(self.enessc*self.e0)))
        cdef double ymax     = self.enessc*self.e0/(self.e0-self.enessc)*self.HC(self.e0-self.enessc) + self.r2*self.HC(self.enessc-self.e0)
        cdef double inty   = scipy.integrate.quad(self.diff_SSC_r,numpy.log10(ymin),numpy.log10(ymax),epsabs=0.0,epsrel=1e-4)[0]
        cdef double SSCv   = inty*self.cnst_v2e*fsync/self.e0**3
        return SSCv

    cpdef double SSCEm(self,double inputv):
        cdef int j
        self.enessc = self.cnst_v2e*inputv
        cdef double c1   = 9.0*pow((1.0+self.z),2.0)*self.sigma*self.enessc**2
        cdef double c2   = 16.0*math.pi*self.D**2*self.c**2*self.T**2
        cdef numpy.ndarray[DTYPE_t,ndim=1] vlst = self.make_log_map(self.vrange_sy0,self.vrange_sy1,self.vssc_step)
        mySEDarray = [ MySED(vlst[j],self.sync[j]) for j in range(self.nsync) ]
        cdef vval = list(map(self.diff_SSC_v,mySEDarray))
        cdef double C0   = scipy.integrate.simps(vval,vlst)
        cdef double SSC  = c1/c2*C0
        return SSC

    ############################################
    ############################################
    ############################################
