import math,sys
import numpy as np
import SSC_calc_cy
from numpy.random import *
import time

NDIM = 9
LOGB, LOGD, LOGT, LOGK, LOGLC, P0, P1, LOGR1, LOGR2 = 0, 1, 2, 3, 4, 5, 6, 7, 8

class SED_Data:
    """
    Dummy class for the argument of logLikelihood().
    """
    def __init__(self,nu,nfn,err,nu_sync_ssc,ssc):
        self.nu = nu
        self.nfn = nfn
        self.err = err
        self.nu_sync_ssc = nu_sync_ssc
        self.ssc = ssc

def logLikelihood(sed):
    """
    Calculate log-likelihood for a given nu.
    """
    if sed.nu <= sed.nu_sync_ssc:
        return (-(sed.nfn-sed.ssc.SyncEm(sed.nu))**2)/(2*sed.err**2)
    else:
        return (-(sed.nfn-sed.ssc.SSCEm(sed.nu))**2)/(2*sed.err**2)

def yysscfit(infile,usecols,n_iteration,z,
             prior_var=None, prior_mu=None, prior_sig=None,
             fixed_var=None, fixed_value=None,
             eed_single_powerlaw=False,
             prm_init=[np.log10(0.001),np.log10(20),np.log10(1361664),np.log10(1e43),np.log10(10**4.5),2.0,4.0,np.log10(10**2),np.log10(10**8)],
             cov_init=[0.04,0.008,0.02,0.6,0.15,0.2,0.2,4.0,7.0],
             nu_sync_ssc=1e19, max_p1=20.0, logr1_fix=2.0, logr2_fix=8.0, random_seed=None):
    """
    Generate the MCMC samples of the SSC parameters.

    Parameters
    ----------
    infile : str
        Name of the input data file including nu, nu_f_nu, and errors.
    usecols : list
        List of the column indexes (zero-based) for nu, nu_f_nu, and errors. 
        If the 1st, 2nd, and 3rd columns in "infile" corresponds to 
        nu, nu_f_nu, and errors, then usecols=[0,1,2].
    n_iteration : int
        The number of iterations of MCMC. 
        Small number (10--100) should be set for a test.
        A full estimaten requires typically, n_iteration=100000.
    z : float
        Redshift
    prior_var : str, optional
        Variable of which Gaussian prior is set. 
        It should be either 'logB', 'logD', 'logT', or 'logK'.
        If it is set, you have to give values of prior_mu and prior_sig.
        (Default: None)
    prior_mu : float, optional
        The mean of the Gaussian prior in log scale.
    prior_sig : float, optional
        The standard deviation of the Gaussian prior in log scale.
    fixed_var : str, optional
        Variable fixed. 
        It should be either 'logB', 'logD', 'logT', or 'logK'.
        If it is set, you have to give values of fixed_value.
        (Default: None)
    fixed_value : float, optional
        The value for the fixed variable.
    eed_single_powerlaw : bool, optional
        If this flag is True, then the model of the electron energy 
        distribution is a single power-law form with a large p1 value 
        of "max_p1". Otherwise, it is a broken power-law form. 
        (Default: False)
    prm_init : list, optional
        Initial SSC parameters. The list should include the 9 variables 
        like, [logB, logD, logT, logK, logLC, p0, p1, logr1, logr2]. 
    cov_init : list, optional
        Initial values of variances for the covariant matrix of 
        the proposal distribution for MCMC. As in "prm_init", it should 
        be a list having 9 values.
    nu_sync_ssc : float, optional
        Frequency below which the synchrotron emission is dominant.
        This code assumes it is dominant below nu_sync_ssc and 
        the SSC is dominant above nu_sync_ssc. It calculates 
        not both components, but either dominant component at each 
        frequency.  (Dafault: 1e19)
    max_p1 : float, optional
        Maximum value of p1.  (Default: 20.0)
    logr1_fix : float, optional
        Fixed value of log gamma_min. (Default: 2.0)
    logr2_fix : float, optional
        Fixed value of log gamma_max. (Default: 8.0)
    random_seed : int, optional
        Ramdom seed number. If a value is set, reproductive MCMC samples 
        are obtained.  (Default: None)
    
    Returns
    -------
    This function only gives outputs to stdout. 
    It outputs a set of SSC parameters for each MCMC iteration.
    It has 12 columns:
    Iteration number, B, D, T, K, LC, p0, p1, r1, r2, acceptance ratio, log likelihood

    """
    if random_seed is not None:
        np.random.seed(seed=random_seed)
    # Inputs
    nu,nfn,err=np.loadtxt(infile,usecols=usecols,unpack=True)
    n_data = len(nu)
    n_iteration = n_iteration
    z = z
    prm0 = np.array(prm_init)
    # Check options
    if prior_var is not None and fixed_var is not None and prior_var is fixed_var:
        raise Exception('prior_var should not be equal to fixed_var.')
    if prior_var is not None and prior_mu is not None and prior_sig is not None:
        flag_prior = True
        if prior_var=='logB':
            index_prior = LOGB
        elif prior_var=='logD':
            index_prior = LOGD
        elif prior_var=='logT':
            index_prior = LOGT
        elif prior_var=='logK':
            index_prior = LOGK
        else:
            raise Exception('prior_var should be "logB", "logD", "logT", or "logK".')
    elif prior_var is not None:
        raise Exception('Values should be given to prior_mu and prior_sig.')
    else:
        flag_prior = False
    if fixed_var is not None and fixed_value is not None:
        flag_fixed = True
        if fixed_var=='logB':
            index_fixed = LOGB
        elif fixed_var=='logD':
            index_fixed = LOGD
        elif fixed_var=='logT':
            index_fixed = LOGT
        elif fixed_var=='logK':
            index_fixed = LOGK
        else:
            raise Exception('fixed_var should be "logB", "logD", "logT", or "logK".')
    elif fixed_var is not None:
        raise Exception('Values should be given to fixed_value.')
    else:
        flag_fixed = False
    # Log posterior for the initial parameters
    ssc = SSC_calc_cy.SSC()
    ssc.set_parameters(z,10**prm0[LOGB],10**prm0[LOGD],10**prm0[LOGT],10**prm0[LOGK],10**prm0[LOGLC],prm0[P0],prm0[P1],10**prm0[LOGR1],10**prm0[LOGR2])
    myArray = [ SED_Data(nu[j],nfn[j],err[j],nu_sync_ssc,ssc) for j in range(n_data) ]
    ### Python 2.x
    # logLi = map(logLikelihood,myArray)
    ### Python 3
    logLi = list(map(logLikelihood,myArray))
    if prior_var is not None :
        logPost = np.sum(logLi) + (-(prm0[index_prior]-prior_mu)**2)/(2*prior_sig**2)
    else:
        logPost = np.sum(logLi)
    # Setup for the adaptive Metropolis
    mu_prev = prm0
    cov_prev = np.diag(cov_init)
    sig_prev = 1
    stp = np.random.multivariate_normal(np.zeros(NDIM), sig_prev*cov_prev,1)
    prm_i = prm0 + stp
    fa = 0
    # Main loop of the adaptive Metropolis
    rej = 0
    irange = range(1,n_iteration+1)
    t1 = time.time()
    for i in irange:
        # Revision of the proposal distribution
        mu_i = mu_prev + (10.0/(i + 100.0))*(prm_i - mu_prev)
        cov_i = cov_prev + (10.0/(i+100.0))*((prm_i - mu_prev)*(prm_i - mu_prev).transpose() - cov_prev)
        sig_i = sig_prev + (10.0/(i+100.0))*(fa - 0.243)
        if i<2:
            prm_i = prm0
        # Create a candidate
        stp = np.random.multivariate_normal(np.zeros(NDIM), sig_i*cov_i,1)
        stp = np.reshape(stp,(NDIM,))
        cand = prm_i + stp
        # Fixed parameters
        if flag_fixed :
            cand[index_fixed] = fixed_value
        if eed_single_powerlaw :
            cand[P1] = max_p1
        cand[LOGR1] = logr1_fix
        cand[LOGR2] = logr2_fix
        # Calculation of the log posterior of the candidate
        if cand[LOGLC] > cand[LOGR1] and cand[LOGR2] > cand[LOGLC] and cand[P1] <= max_p1 and cand[LOGD] < 40.0:
            ssc.set_parameters(z,10**cand[LOGB],10**cand[LOGD],10**cand[LOGT],10**cand[LOGK],10**cand[LOGLC],cand[P0],cand[P1],10**cand[LOGR1],10**cand[LOGR2])
            myArray = [ SED_Data(nu[j],nfn[j],err[j],nu_sync_ssc,ssc) for j in range(n_data) ]
            ### Python 2.x
            # logLi_c = map(logLikelihood,myArray)
            ### Python 3
            logLi_c = list(map(logLikelihood,myArray))
            if flag_prior :
                logPost_c = np.sum(logLi_c) + (-(cand[index_prior]-prior_mu)**2)/(2*prior_sig**2)
            else:
                logPost_c = np.sum(logLi_c)
        else:
            logPost_c = -1.0e+300
        # Decision of the next state by the Metropolis algorithm
        log_ratio = logPost_c-logPost
        if log_ratio > 0:
            prm_i = cand
            fa = 1
            logPost = logPost_c
        else:
            if log_ratio >= math.log(np.random.rand()):
                prm_i = cand
                fa = 1
                logPost = logPost_c
            else:
                rej +=1
                fa = 0
        # Revision of the proposal parameters
        mu_prev = mu_i
        cov_prev = cov_i
        sig_prev = sig_i
        accep =  1 - float(rej)/i
        # Output
        ### Python 2.x
        # print i,10**prm_i[LOGB],10**prm_i[LOGD],10**prm_i[LOGT],10**prm_i[LOGK],10**prm_i[LOGLC],prm_i[P0],prm_i[P1],10**prm_i[LOGR1],10**prm_i[LOGR2],accep,logPost
        ### Python 3
        print(i,10**prm_i[LOGB],10**prm_i[LOGD],10**prm_i[LOGT],10**prm_i[LOGK],10**prm_i[LOGLC],prm_i[P0],prm_i[P1],10**prm_i[LOGR1],10**prm_i[LOGR2],accep,logPost)
    t2 = time.time()
    sys.stderr.write("Time: %f sec (%d steps)\n" % (t2-t1,i))
